﻿#include <cassert>
#include <chrono> // 'std::chrono::duration'...
#include <filesystem> // 'std::filesystem::path'...
#include <fstream> // 'std::ifstream'...
#include <iostream>
#include <string> // 'std::getline'...
#include <string_view>

#include "GTIN.h"

std::ifstream::event_callback call_back_function = [](std::ios_base::event event, std::ios_base& stream, int index) -> void {
	if (event == std::ios_base::erase_event)
		std::cout << "std::ios_base::erase_event" << std::endl;
	};

std::filesystem::path const GTIN::Path = std::filesystem::path("/Users/franc/Desktop/GTIN/GTIN.csv");

std::list<std::string_view> GTIN::_Slice_line(const std::string& line) {
	std::list<std::string_view> sliced_line;
	std::string::const_iterator from = line.end(); // Invalid position...
	std::string::const_iterator to = line.end(); // Invalid position...
	for (std::string::const_iterator i = line.begin(); !(i == line.end()); i++) {
		to = *i == '"' && from != line.end() ? i : to;
		from = *i == '"' && from == line.end() ? i : from;
		if (from != line.end() && to != line.end()) {
			sliced_line.push_back(std::string_view{ ++from, to }); // Copy, but cheap...
			// Next round...
			from = line.end();
			to = line.end();
		}
	}
	return sliced_line; // Copy, but cheap...
}

GTIN::GTIN(const std::string& file_name) {
	std::ifstream ifstream;
	if (file_name != "")
		ifstream.open(file_name/*, std::ios_base::in*/);
	else
		ifstream.open(GTIN::Path/*, std::ios_base::in*/);
	if (not ifstream.is_open())
		throw std::ios::failure("Failed to open file....");
	assert(ifstream.is_open());
	ifstream.register_callback(call_back_function, 0); // Call back function on stream ('0' is custom data)...

	/* First line */
	std::string line;
	std::getline(ifstream, line, '\r'); // Windows!
	std::list<std::string_view> sliced_line = GTIN::_Slice_line(line);
	// for (auto& slice : sliced_line) std::cout << slice << '\t';

	std::chrono::time_point now = std::chrono::high_resolution_clock::now(); // System clock...
	while (std::getline(ifstream, line, '\r')) { // Fill in memory with data inside OS buffer...
		std::string line_copy = line; // Copy, expensive...
		_lines.push_back(line_copy); // Copy, expensive...
		// Bug, string views refer to positions in 'line_copy', which soon vanishes:
		// _data.push_back(GTIN::_Slice_line(line_copy));
		_data.push_back(GTIN::_Slice_line(_lines.back())); // Bug suppression...

		// _lines.push_back(std::move(line)); // Move, line is "consumed"...
		// _data.push_back(GTIN::_Slice_line(_lines.back()));
	}
	// 'std::chrono::milliseconds' 1s / 1 000
	// 'std::chrono::microseconds' 1s / 1 000 000
	// 'std::chrono::nanoseconds' 1s / 1 000 000 000
	std::chrono::duration duration = std::chrono::duration_cast<std::chrono::microseconds>(std::chrono::high_resolution_clock::now() - now);
	std::cout << "Elapsed time: " << duration.count() << std::endl << std::endl;
	
	ifstream.close(); // Useless, done automatically...

	for (auto& datum : _data) { for (auto& item : datum) std::cout << item << '\t'; std::cout << std::endl << std::endl; }
}

int main() {
	// std::cout << "Path of working directory: " << std::filesystem::current_path() << std::endl;
	GTIN{};

	return 0;
}
